@@ -12,14 +12,16 @@ import email.policy
 import requests
 import urllib.parse
 import datetime
+import time
 
+from pathlib import Path
 from tempfile import mkstemp
 
 from email import charset
 charset.add_charset('utf-8', None)
 emlpolicy = email.policy.EmailPolicy(utf8=True, cte_type='8bit', max_line_length=None)
 
-VERSION = '0.3.3'
+__VERSION__ = '0.3.5'
 ATTESTATION_FORMAT_VER = '0.1'
 
 logger = logging.getLogger('b4')
@@ -84,6 +86,8 @@ DEFAULT_CONFIG = {
     'attestation-gnupghome': None,
     # Do you like simple or fancy checkmarks?
     'attestation-checkmarks': 'fancy',
+    # How long to keep things in cache before expiring (minutes)?
+    'cache-expire': '10',
     # If this is not set, we'll use what we find in 
     # git-config for gpg.program, and if that's not set,
     # we'll use "gpg" and hope for the better
@@ -101,12 +105,15 @@ ATTESTATIONS = list()
 SUBKEY_DATA = dict()
 # Used for storing our requests session
 REQSESSION = None
+# Indicates that we've cleaned cache already
+_CACHE_CLEANED = False
 
 
 class LoreMailbox:
     def __init__(self):
         self.msgid_map = dict()
         self.series = dict()
+        self.covers = dict()
         self.followups = list()
         self.unknowns = list()
 
@@ -128,7 +135,7 @@ class LoreMailbox:
             return self.msgid_map[msgid]
         return None
 
-    def get_series(self, revision=None):
+    def get_series(self, revision=None, sloppytrailers=False):
         if revision is None:
             if not len(self.series):
                 return None
@@ -149,8 +156,11 @@ class LoreMailbox:
             logger.critical('All patches in series v%s are missing.', lser.revision)
             return None
 
-        # Do we have a cover letter for it?
-        if not lser.has_cover:
+        # Grab our cover letter if we have one
+        if revision in self.covers.keys():
+            lser.add_patch(self.covers[revision])
+            lser.has_cover = True
+        else:
             # Let's find the first patch with an in-reply-to and see if that
             # is our cover letter
             for member in lser.patches:
@@ -191,12 +201,14 @@ class LoreMailbox:
             else:
                 pmsg = self.msgid_map[fmsg.in_reply_to]
 
-            trailers = fmsg.trailers
+            trailers, mismatches = fmsg.get_trailers(sloppy=sloppytrailers)
+            for tname, tvalue in mismatches:
+                lser.trailer_mismatches.add((tvalue, fmsg.fromname, fmsg.fromemail))
             lvl = 1
             while True:
                 logger.debug('%sParent: %s', ' ' * lvl, pmsg.full_subject)
                 logger.debug('%sTrailers:', ' ' * lvl)
-                for trailer in set(trailers):
+                for trailer in trailers:
                     logger.debug('%s%s: %s', ' ' * (lvl+1), trailer[0], trailer[1])
                 found = False
                 if lser.revision != pmsg.revision:
@@ -204,14 +216,14 @@ class LoreMailbox:
                 for lmsg in lser.patches:
                     if lmsg is not None and lmsg.msgid == pmsg.msgid:
                         # Confirmed, this is our parent patch
-                        lmsg.followup_trailers += trailers
+                        lmsg.followup_trailers.update(trailers)
                         found = True
                         break
                 if found:
                     break
                 elif pmsg.in_reply_to and pmsg.in_reply_to in self.msgid_map:
                     lvl += 1
-                    trailers += pmsg.trailers
+                    trailers.update(pmsg.trailers)
                     pmsg = self.msgid_map[pmsg.in_reply_to]
                 else:
                     break
@@ -223,42 +235,58 @@ class LoreMailbox:
         logger.debug('Looking at: %s', lmsg.full_subject)
         self.msgid_map[lmsg.msgid] = lmsg
 
-        if lmsg.has_diff or lmsg.has_diffstat:
+        if lmsg.counter == 0 and lmsg.has_diffstat:
+            # Cover letter
+            # Add it to covers -- we'll deal with them later
+            logger.debug('  adding as v%s cover letter', lmsg.revision)
+            self.covers[lmsg.revision] = lmsg
+            return
+
+        if lmsg.reply:
+            # We'll figure out where this belongs later
+            logger.debug('  adding to followups')
+            self.followups.append(lmsg)
+            return
+
+        if re.search(r'^Comment: att-fmt-ver:', lmsg.body, re.I | re.M):
+            logger.debug('Found attestation message')
+            LoreAttestationDocument.load_from_string(lmsg.msgid, lmsg.body)
+            # We don't keep it, because it's not useful for us beyond this point
+            return
+
+        if lmsg.has_diff:
+            if lmsg.revision not in self.series:
+                if lmsg.revision_inferred and lmsg.in_reply_to:
+                    # We have an inferred revision here.
+                    # Do we have an upthread cover letter that specifies a revision?
+                    irt = self.get_by_msgid(lmsg.in_reply_to)
+                    if irt is not None and irt.has_diffstat and not irt.has_diff:
+                        # Yes, this is very likely our cover letter
+                        logger.debug('  fixed revision to v%s', irt.revision)
+                        lmsg.revision = irt.revision
+
+            # Run our check again
             if lmsg.revision not in self.series:
                 self.series[lmsg.revision] = LoreSeries(lmsg.revision, lmsg.expected)
                 if len(self.series) > 1:
                     logger.info('Found new series v%s', lmsg.revision)
-            if lmsg.has_diff:
-                # Attempt to auto-number series from the same author who did not bother
-                # to set v2, v3, etc in the patch revision
-                if (lmsg.counter == 1 and lmsg.counters_inferred
-                        and not lmsg.reply and lmsg.lsubject.patch and not lmsg.lsubject.resend):
-                    omsg = self.series[lmsg.revision].patches[lmsg.counter]
-                    if (omsg is not None and omsg.counters_inferred and lmsg.fromemail == omsg.fromemail
-                            and omsg.date < lmsg.date):
-                        lmsg.revision = len(self.series) + 1
-                        self.series[lmsg.revision] = LoreSeries(lmsg.revision, lmsg.expected)
-                        logger.info('Assuming new revision: v%s (%s)', lmsg.revision, lmsg.full_subject)
-                logger.debug('  adding as patch')
-                self.series[lmsg.revision].add_patch(lmsg)
-            elif lmsg.counter == 0 and lmsg.has_diffstat:
-                # Bona-fide cover letter
-                logger.debug('  adding as cover letter')
-                self.series[lmsg.revision].add_cover(lmsg)
-            elif lmsg.reply:
-                # We'll figure out where this belongs later
-                logger.debug('  adding to followups')
-                self.followups.append(lmsg)
-        elif lmsg.reply:
-            logger.debug('  adding to followups')
-            self.followups.append(lmsg)
-        elif re.search(r'^Comment: att-fmt-ver:', lmsg.body, re.I | re.M):
-            logger.debug('Found attestation message')
-            LoreAttestationDocument.load_from_string(lmsg.msgid, lmsg.body)
-            # We don't keep it, because it's not useful for us beyond this point
-        else:
-            logger.debug('  adding to unknowns')
-            self.unknowns.append(lmsg)
+
+            # Attempt to auto-number series from the same author who did not bother
+            # to set v2, v3, etc in the patch revision
+            if (lmsg.counter == 1 and lmsg.counters_inferred
+                    and not lmsg.reply and lmsg.lsubject.patch and not lmsg.lsubject.resend):
+                omsg = self.series[lmsg.revision].patches[lmsg.counter]
+                if (omsg is not None and omsg.counters_inferred and lmsg.fromemail == omsg.fromemail
+                        and omsg.date < lmsg.date):
+                    lmsg.revision = len(self.series) + 1
+                    self.series[lmsg.revision] = LoreSeries(lmsg.revision, lmsg.expected)
+                    logger.info('Assuming new revision: v%s (%s)', lmsg.revision, lmsg.full_subject)
+            logger.debug('  adding as patch')
+            self.series[lmsg.revision].add_patch(lmsg)
+            return
+
+        logger.debug('  adding to unknowns')
+        self.unknowns.append(lmsg)
 
 
 class LoreSeries:
@@ -267,6 +295,7 @@ class LoreSeries:
         self.expected = expected
         self.patches = [None] * (expected+1)
         self.followups = list()
+        self.trailer_mismatches = set()
         self.complete = False
         self.has_cover = False
 
@@ -311,11 +340,7 @@ class LoreSeries:
             self.patches[lmsg.counter] = lmsg
         self.complete = not (None in self.patches[1:])
 
-    def add_cover(self, lmsg):
-        self.add_patch(lmsg)
-        self.has_cover = True
-
-    def get_slug(self):
+    def get_slug(self, extended=False):
         # Find the first non-None entry
         lmsg = None
         for lmsg in self.patches:
@@ -327,8 +352,14 @@ class LoreSeries:
 
         prefix = lmsg.date.strftime('%Y%m%d')
         authorline = email.utils.getaddresses(lmsg.msg.get_all('from', []))[0]
-        author = re.sub(r'\W+', '_', authorline[1]).strip('_').lower()
-        slug = '%s_%s' % (prefix, author)
+        if extended:
+            local = authorline[1].split('@')[0]
+            unsafe = '%s_%s_%s' % (prefix, local, lmsg.subject)
+            slug = re.sub(r'\W+', '_', unsafe).strip('_').lower()
+        else:
+            author = re.sub(r'\W+', '_', authorline[1]).strip('_').lower()
+            slug = '%s_%s' % (prefix, author)
+
         if self.revision != 1:
             slug = 'v%s_%s' % (self.revision, slug)
 
@@ -367,11 +398,11 @@ class LoreSeries:
         for lmsg in self.patches[1:]:
             if lmsg is not None:
                 if self.has_cover and covertrailers and self.patches[0].followup_trailers:
-                    lmsg.followup_trailers += self.patches[0].followup_trailers
+                    lmsg.followup_trailers.update(self.patches[0].followup_trailers)
                 if addmysob:
-                    lmsg.followup_trailers.append(('Signed-off-by', '%s <%s>' % (usercfg['name'], usercfg['email'])))
+                    lmsg.followup_trailers.add(('Signed-off-by', '%s <%s>' % (usercfg['name'], usercfg['email'])))
                 if addlink:
-                    lmsg.followup_trailers.append(('Link', linkmask % lmsg.msgid))
+                    lmsg.followup_trailers.add(('Link', linkmask % lmsg.msgid))
 
                 if attpolicy != 'off':
                     lore_lookup = False
@@ -482,8 +513,8 @@ class LoreMessage:
         self.body = None
         self.has_diff = False
         self.has_diffstat = False
-        self.trailers = list()
-        self.followup_trailers = list()
+        self.trailers = set()
+        self.followup_trailers = set()
 
         self.attestation = None
 
@@ -541,6 +572,12 @@ class LoreMessage:
             if diffre.search(payload):
                 self.body = payload
 
+        if self.body is None:
+            # Woah, we didn't find any usable parts
+            logger.debug('  No plain or patch parts found in message')
+            logger.info('  Not plaintext: %s', self.full_subject)
+            return
+
         if diffstatre.search(self.body):
             self.has_diffstat = True
         if diffre.search(self.body):
@@ -552,52 +589,68 @@ class LoreMessage:
             matches = re.findall(r'^\s*Fixes:[ \t]+([a-f0-9]+\s+\(.*\))\s*$', self.body, re.MULTILINE)
             if matches:
                 for tvalue in matches:
-                    self.trailers.append(('Fixes', tvalue))
+                    self.trailers.add(('Fixes', tvalue))
 
             # Do we have something that looks like a person-trailer?
             matches = re.findall(r'^\s*([\w-]+):[ \t]+(.*<\S+>)\s*$', self.body, re.MULTILINE)
+            # These are commonly part of patch/commit metadata
+            badtrailers = ('from', 'author', 'email')
             if matches:
-                # Basic sanity checking -- the trailer must match the name or the email
-                # in the From header, to avoid false-positive trailer parsing errors
                 for tname, tvalue in matches:
-                    tmatch = False
-                    namedata = email.utils.getaddresses([tvalue])[0]
-                    tfrom = re.sub(r'\+[^@]+@', '@', namedata[1].lower())
-                    hfrom = re.sub(r'\+[^@]+@', '@', self.fromemail.lower())
-                    tlname = namedata[0].lower()
-                    hlname = self.fromname.lower()
-                    tchunks = tfrom.split('@')
-                    hchunks = hfrom.split('@')
-                    if tfrom == hfrom:
-                        logger.debug('  trailer exact email match')
-                        tmatch = True
-                    # See if domain part of one of the addresses is a subset of the other one,
-                    # which should match cases like @linux.intel.com and @intel.com
-                    elif (len(tchunks) == 2 and len(hchunks) == 2
-                          and tchunks[0] == hchunks[0]
-                          and (tchunks[1].find(hchunks[1]) >= 0 or hchunks[1].find(tchunks[1]) >= 0)):
-                        logger.debug('  trailer fuzzy email match')
-                        tmatch = True
-                    # Does the name match, at least?
-                    elif tlname == hlname:
-                        logger.debug('  trailer exact name match')
-                        tmatch = True
-                    # Finally, see if the header From has a comma in it and try to find all
-                    # parts in the trailer name
-                    elif hlname.find(',') > 0:
-                        nmatch = True
-                        for nchunk in hlname.split(','):
-                            if hlname.find(nchunk.strip()) < 0:
-                                nmatch = False
-                                break
-                        if nmatch:
-                            logger.debug('  trailer fuzzy name match')
-                            tmatch = True
-                    if tmatch:
-                        self.trailers.append((tname, tvalue))
-                    else:
-                        logger.debug('  ignoring "%s: %s" due to from mismatch (from: %s %s)', tname, tvalue,
-                                     self.fromname, self.fromemail)
+                    if tname.lower() not in badtrailers:
+                        self.trailers.add((tname, tvalue))
+
+    def get_trailers(self, sloppy=False):
+        mismatches = set()
+        if sloppy:
+            return set(self.trailers), mismatches
+
+        trailers = set()
+        for tname, tvalue in self.trailers:
+            # Don't do from-matching on certain trailers like Fixes:
+            if tname.lower() in ('fixes',):
+                trailers.add((tname, tvalue))
+                continue
+
+            tmatch = False
+            namedata = email.utils.getaddresses([tvalue])[0]
+            tfrom = re.sub(r'\+[^@]+@', '@', namedata[1].lower())
+            hfrom = re.sub(r'\+[^@]+@', '@', self.fromemail.lower())
+            tlname = namedata[0].lower()
+            hlname = self.fromname.lower()
+            tchunks = tfrom.split('@')
+            hchunks = hfrom.split('@')
+            if tfrom == hfrom:
+                logger.debug('  trailer exact email match')
+                tmatch = True
+            # See if domain part of one of the addresses is a subset of the other one,
+            # which should match cases like @linux.intel.com and @intel.com
+            elif (len(tchunks) == 2 and len(hchunks) == 2
+                  and tchunks[0] == hchunks[0]
+                  and (tchunks[1].find(hchunks[1]) >= 0 or hchunks[1].find(tchunks[1]) >= 0)):
+                logger.debug('  trailer fuzzy email match')
+                tmatch = True
+            # Does the name match, at least?
+            elif tlname == hlname:
+                logger.debug('  trailer exact name match')
+                tmatch = True
+            # Finally, see if the header From has a comma in it and try to find all
+            # parts in the trailer name
+            elif hlname.find(',') > 0:
+                nmatch = True
+                for nchunk in hlname.split(','):
+                    if hlname.find(nchunk.strip()) < 0:
+                        nmatch = False
+                        break
+                if nmatch:
+                    logger.debug('  trailer fuzzy name match')
+                    tmatch = True
+            if tmatch:
+                trailers.add((tname, tvalue))
+            else:
+                mismatches.add((tname, tvalue))
+
+        return trailers, mismatches
 
     def __repr__(self):
         out = list()
@@ -630,8 +683,7 @@ class LoreMessage:
 
     @staticmethod
     def clean_header(hdrval):
-        uval = hdrval.replace('\n', ' ')
-        new_hdrval = re.sub(r'\s+', ' ', uval)
+        new_hdrval = re.sub(r'\n?\s+', ' ', str(hdrval))
         return new_hdrval.strip()
 
     @staticmethod
@@ -654,7 +706,6 @@ class LoreMessage:
         # XXX: This currently doesn't work for git binary patches
         #
         diff = diff.replace('\r', '')
-        diff = diff.strip() + '\n'
 
         # For keeping a buffer of lines preceding @@ ... @@
         buflines = list()
@@ -685,8 +736,9 @@ class LoreMessage:
             if pp > 0:
                 # Inside the patch
                 phasher.update((line + '\n').encode('utf-8'))
-                if line[0] != '-':
-                    pp -= 1
+                if len(line) and line[0] == '-':
+                    continue
+                pp -= 1
                 continue
             # Not anything we recognize, so stick into buflines
             buflines.append(line)
@@ -755,7 +807,8 @@ class LoreMessage:
 
                 # Now we add mix-in trailers
                 btrailers.reverse()
-                trailers = btrailers + self.followup_trailers
+                trailers = set(self.followup_trailers)
+                trailers.update(set(btrailers))
                 added = list()
                 if trailer_order is None:
                     trailer_order = DEFAULT_TRAILER_ORDER
@@ -766,9 +819,9 @@ class LoreMessage:
                         if fnmatch.fnmatch(trailer[0].lower(), trailermatch.strip()):
                             fixlines.append('%s: %s' % trailer)
                             if trailer not in btrailers:
-                                logger.info('    Added: %s: %s' % trailer)
+                                logger.info('    + %s: %s' % trailer)
                             else:
-                                logger.debug('     Kept: %s: %s' % trailer)
+                                logger.debug('    . %s: %s' % trailer)
                             added.append(trailer)
                 trailersdone = True
             fixlines.append(line)
@@ -851,8 +904,8 @@ class LoreSubject:
         # Remove any brackets inside brackets
         while True:
             oldsubj = subject
-            subject = re.sub(r'^\s*\[([^\]]*)\[([^\]]*)\]', '[\\1\\2]', subject)
-            subject = re.sub(r'^\s*\[([^\]]*)\]([^\]]*)\]', '[\\1\\2]', subject)
+            subject = re.sub(r'^\s*\[([^\]]*)\[([^\[\]]*)\]', '[\\1\\2]', subject)
+            subject = re.sub(r'^\s*\[([^\]]*)\]([^\[\]]*)\]', '[\\1\\2]', subject)
             if oldsubj == subject:
                 break
 
@@ -1016,6 +1069,15 @@ class LoreAttestationDocument:
         if self.good and self.valid and self.trusted:
             self.passing = True
 
+        if source.find('http') == 0:
+            # We only cache known-good attestations obtained from remote
+            cachedir = get_cache_dir()
+            cachename = '%s.attestation' % urllib.parse.quote_plus(source.strip('/').split('/')[-1])
+            fullpath = os.path.join(cachedir, cachename)
+            with open(fullpath, 'w') as fh:
+                logger.debug('Saved attestation in cache: %s', cachename)
+                fh.write(sigdata)
+
         hg = [None, None, None]
         for line in sigdata.split('\n'):
             # It's a yaml document, but we don't parse it as yaml for safety reasons
@@ -1052,26 +1114,60 @@ class LoreAttestationDocument:
             out.append('  | %s-%s-%s' % (hg[0][:8], hg[1][:8], hg[2][:8]))
         return '\n'.join(out)
 
+    @staticmethod
+    def get_from_cache(attid):
+        cachedir = get_cache_dir()
+        attdocs = list()
+        for entry in os.listdir(cachedir):
+            if entry.find('.attestation') <= 0:
+                continue
+            fullpath = os.path.join(cachedir, entry)
+            with open(fullpath, 'r') as fh:
+                content = fh.read()
+                # Can't be 0, because it has to have pgp ascii wrapper
+                if content.find(attid) > 0:
+                    attdoc = LoreAttestationDocument(fullpath, content)
+                    attdocs.append(attdoc)
+        return attdocs
+
     @staticmethod
     def get_from_lore(attid):
+        attdocs = list()
         # XXX: Querying this via the Atom feed is a temporary kludge until we have
         #      proper search API on lore.kernel.org
+        cachedir = get_cache_dir()
+        cachefile = os.path.join(cachedir, '%s.lookup' % urllib.parse.quote_plus(attid))
+        status = None
+        if os.path.exists(cachefile):
+            with open(cachefile, 'r') as fh:
+                try:
+                    status = int(fh.read())
+                except ValueError:
+                    pass
+        if status is not None and status != 200:
+            logger.debug('Cache says looking up %s = %s', attid, status)
+            return attdocs
+
         config = get_main_config()
         queryurl = '%s?%s' % (config['attestation-query-url'],
                               urllib.parse.urlencode({'q': attid, 'x': 'A', 'o': '-1'}))
         logger.debug('Query URL: %s', queryurl)
         session = get_requests_session()
         resp = session.get(queryurl)
-        content = resp.content.decode('utf-8')
+        if resp.status_code != 200:
+            # Record this as a bad hit
+            with open(cachefile, 'w') as fh:
+                fh.write(str(resp.status_code))
+
         matches = re.findall(
             r'link\s+href="([^"]+)".*?(-----BEGIN PGP SIGNED MESSAGE-----.*?-----END PGP SIGNATURE-----)',
-            content, flags=re.DOTALL
+            resp.content.decode('utf-8'), flags=re.DOTALL
         )
 
-        attdocs = list()
         if matches:
             for link, sigdata in matches:
-                attdocs.append(LoreAttestationDocument(link, sigdata))
+                attdoc = LoreAttestationDocument(link, sigdata)
+                attdocs.append(attdoc)
 
         return attdocs
 
@@ -1097,22 +1193,27 @@ class LoreAttestation:
         self.passing = False
         self.attdocs = list()
 
-    def validate(self, lore_lookup=True):
+    def _check_if_passing(self):
         global ATTESTATIONS
-
         hg = (self.i, self.m, self.p)
         for attdoc in ATTESTATIONS:
             if hg in attdoc.hashes and attdoc.passing:
                 self.passing = True
                 self.attdocs.append(attdoc)
 
+    def validate(self, lore_lookup=True):
+        global ATTESTATIONS
+        self._check_if_passing()
+
+        if not len(self.attdocs):
+            attdocs = LoreAttestationDocument.get_from_cache(self.attid)
+            ATTESTATIONS += attdocs
+            self._check_if_passing()
+
         if not len(self.attdocs) and lore_lookup:
             attdocs = LoreAttestationDocument.get_from_lore(self.attid)
             ATTESTATIONS += attdocs
-            for attdoc in attdocs:
-                if hg in attdoc.hashes and attdoc.passing:
-                    self.passing = True
-                    self.attdocs.append(attdoc)
+            self._check_if_passing()
 
     def __repr__(self):
         out = list()
@@ -1212,6 +1313,36 @@ def get_main_config():
     return MAIN_CONFIG
 
 
+def get_cache_dir():
+    global _CACHE_CLEANED
+    if 'XDG_CACHE_HOME' in os.environ:
+        cachehome = os.environ['XDG_CACHE_HOME']
+    else:
+        cachehome = os.path.join(str(Path.home()), '.cache')
+    cachedir = os.path.join(cachehome, 'b4')
+    Path(cachedir).mkdir(parents=True, exist_ok=True)
+    if _CACHE_CLEANED:
+        return cachedir
+
+    # Delete all .mbx and .lookup files older than cache-expire
+    config = get_main_config()
+    try:
+        expmin = int(config['cache-expire']) * 60
+    except ValueError:
+        logger.critical('ERROR: cache-expire must be an integer (minutes): %s', config['cache-expire'])
+        expmin = 600
+    expage = time.time() - expmin
+    for entry in os.listdir(cachedir):
+        if entry.find('.mbx') <= 0 and entry.find('.lookup') <= 0:
+            continue
+        st = os.stat(os.path.join(cachedir, entry))
+        if st.st_mtime < expage:
+            logger.debug('Cleaning up cache: %s', entry)
+            os.unlink(os.path.join(cachedir, entry))
+    _CACHE_CLEANED = True
+    return cachedir
+
+
 def get_user_config():
     global USER_CONFIG
     if USER_CONFIG is None:
@@ -1223,7 +1354,7 @@ def get_requests_session():
     global REQSESSION
     if REQSESSION is None:
         REQSESSION = requests.session()
-        REQSESSION.headers.update({'User-Agent': 'b4/%s' % VERSION})
+        REQSESSION.headers.update({'User-Agent': 'b4/%s' % __VERSION__})
     return REQSESSION
 
 
